from Orange.widgets import widget, gui, settings
from Orange.data import Table
from pathlib import Path
import json
from ..download.FileWatcher import FileWatcher
from ..webserver.start_server import start_server

class BaseLocalInterfaceWidget(widget.OWWidget, openclass=True):
    # Common settings across all widgets
    workflow_id = settings.Setting("default.ows")  # type: ignore
    input_id = settings.Setting("default_id")  # type: ignore

    class Outputs:
        data = widget.Output("Data", Table)

    def __init__(self):
        super().__init__()
        # GUI elements common to all widgets
        self.info_label = gui.label(self.controlArea, self, "Initial info.")
        self.workflow_id_edit = gui.lineEdit(self.controlArea, self, "workflow_id",
                                             label="Workflow ID:", orientation="horizontal")
        self.input_id_edit = gui.lineEdit(self.controlArea, self, "input_id",
                                         label="input ID:", orientation="horizontal")
        
        # Set up file watcher for all widgets
        self.queue_file_path = self.setup_paths()
        self.file_watcher = FileWatcher(str(self.queue_file_path))
        self.file_watcher.file_modified.connect(self.process_queue)
        # Start server for all widgets
        start_server()

    def setup_paths(self) -> Path:
        # Set default path, can be overridden in derived class if needed
        return Path(__file__).parent.parent / "webserver" / "received_files" / "in_process_queue.json"
    
        
        
    
    def get_files_to_process(self, entry) -> list[Path]:
        try:
            self.opening_method
        except AttributeError:
            raise NotImplementedError("Derived classes must give a value for the attribute 'opening_method'.")
        
        if entry.get("workflow_id") != self.workflow_id:
            return []
        
        file_to_process = []
        for file_entry in entry.get("file_entries", []):
            if file_entry.get("file_id") == self.input_id \
                and file_entry.get("opening_method") == self.opening_method:
                file_path = Path(file_entry["input_filename"])
                if file_path.exists():
                    file_to_process.append(file_path)
                else:
                    self.info_label.setText(f"File not found: {str(file_path)}")
                    print("file not found: ", str(file_path))

        return file_to_process
    


    def process_queue(self) -> None:
        # Standard queue processing logic
        if not self.queue_file_path.exists():
            self.info_label.setText("Queue file not found.")
            return

        try:
            with open(self.queue_file_path, 'r') as queue_file:
                content = queue_file.read().strip()
                if not content:
                    self.info_label.setText("Queue file is empty.")
                    return
                files_to_process = json.loads(content)
        except Exception as e:
            self.info_label.setText(f"Error reading queue file: {e}")

        entries_to_process = [entry for entry in files_to_process if entry.get("workflow_id") == self.workflow_id]
        assert len(entries_to_process) <= 1, "Multiple entries found for a single workflow ID."
        if entries_to_process:
            files_to_process = self.get_files_to_process(entries_to_process[0])
            self.process_files(files_to_process)
            

    def process_files(self, files_to_process: list[Path]):
        # To be implemented in derived classes
        raise NotImplementedError("Derived classes from BaseDownloadWidget must implement the process_entry method.")




import hashlib
import uuid
import zipfile
from pathlib import Path
from typing import Tuple


def hash_file_content(content: bytes) -> str:
    """Compute the MD5 hash of a given content."""
    return hashlib.md5(content).hexdigest()

def generate_unique_output_folder(output_directory: Path) -> Tuple[str, Path]:
    """Generate a directory for the outputs using a UUID."""
    unique_id: str = str(uuid.uuid4())
    output_folder: Path = output_directory / unique_id
    output_folder.mkdir(exist_ok=True)
    return unique_id, output_folder

def make_list(liste: list) -> list:
    """Convert a string with comma-separated values to a list."""
    if len(liste) == 1 and "," in liste[0]:
        return liste[0].split(",")
    else:
        return liste
    
def zip_folder(source_dir: Path, output_filename: Path, excluded_files: list[str]=["done", "output.zip"]):
    """Zip the contents of a folder, excluding certain files."""
    if output_filename.exists():
        print("Output file already exists. Skipping zipping.")
        return
    with zipfile.ZipFile(output_filename, 'w', zipfile.ZIP_DEFLATED) as zipf:
         for file_path in source_dir.rglob('*'):
                    if file_path.is_file() and file_path.name not in excluded_files:
                        # Write the file to the archive with an appropriate relative path
                        zipf.write(file_path, file_path.relative_to(source_dir))




import json
import os

from Orange.widgets import widget

from ..download.BaseLocalInterfaceWidget import BaseLocalInterfaceWidget


class OWLocalInterfaceText(BaseLocalInterfaceWidget):
    name = "Local Interface - Text"
    description = "Get textual data from a local interface"
    icon = "icons/local_interf_text_pull.svg"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/local_interf_text_pull.svg"
    priority=1214
    category = "Advanced Artificial Intelligence Tools"
    class Outputs:
        data = widget.Output("data", str)

    def __init__(self):
        super().__init__()
        self.Outputs.data.send("")

    def process_queue(self) -> None:
        # Standard queue processing logic
        if not self.queue_file_path.exists():
            self.info_label.setText("Queue file not found.")
            return

        try:
            with open(self.queue_file_path, 'r') as queue_file:
                content = queue_file.read().strip()
                if not content:
                    self.info_label.setText("Queue file is empty.")
                    return
                files_to_process = json.loads(content)
        except Exception as e:
            self.info_label.setText(f"Error reading queue file: {e}")

        entries_to_process = [entry for entry in files_to_process if entry.get("workflow_id") == self.workflow_id]
        assert len(entries_to_process) <= 1, "Multiple entries found for a single workflow ID."
        if entries_to_process:
            entry = entries_to_process[0]
            text_input = entry.get("text_inputs")
            if len(text_input) % 2 != 0:
                self.info_label.setText("Text input not in pairs.")
                return

            for key, value in zip(text_input[::2], text_input[1::2]):
                if key == self.input_id:
                    self.info_label.setText(f"Text input found: {value}")
                    self.Outputs.data.send(value)
                    break
            




import json
from pathlib import Path
from typing import Any, Dict, List


class FileQueue:
    def __init__(self, file_path: Path):
        self.file_path: Path = file_path # path to json file to be processed
        self.queue: List[Dict[str, Any]] = []
        self.file_path.parent.mkdir(parents=True, exist_ok=True)
        self.file_path.touch(exist_ok=True)

    def load_queue(self) -> List[Dict[str, Any]]:
        """Load queue from the file, or initialize an empty queue if the file does not exist."""
        if self.file_path.exists():
            with open(self.file_path, 'r') as file:
                return json.load(file)
        else:
            return []

    def save_queue(self) -> None:
        """Persist the current state of the queue to a file."""
        with open(self.file_path, 'w') as file:
            json.dump(self.queue, file)

    def add(self, item: Dict[str, Any]) -> None:
        """Add an item to the queue and save the updated queue."""
        self.queue.append(item)
        self.save_queue()

    def remove(self, item: Dict[str, Any]) -> None:
        """Remove an item from the queue and save the updated queue."""
        if item in self.queue:
            self.queue.remove(item)
            self.save_queue()
        else:
            print(f"Attempted to remove non-existent item from queue: {item}")


import os
import socket
import subprocess
import sys
import threading
from logging import getLogger
from pathlib import Path
from typing import Literal

from uvicorn import Config, Server

logger = getLogger(__name__)

def is_port_in_use(port: int) -> bool:
    """Check if a port is already in use."""
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(('127.0.0.1', port)) == 0
    

def start_server(port: int = 8000, method: Literal["new_terminal", "same_terminal"] = "new_terminal") -> None:
    print("Start server called with method: ", method)
    if is_port_in_use(port):
        print("Server is already running.")
        return
    if method == "new_terminal":
        _start_server_in_new_terminal(port)
    elif method == "same_terminal":
        _start_server_in_same_terminal(port)
    else:
        print(f"Invalid method: {method}")


def _start_server_in_new_terminal(port: int = 8000):
    """Starts the server in a new terminal."""
    path_uvicorn="uvicorn"
    if os.name == "nt":
        path_uvicorn=os.path.dirname(sys.executable)
        path_uvicorn=path_uvicorn.replace("\\","/")
        path_uvicorn+="/Scripts/uvicorn.exe"
        if not os.path.isfile(path_uvicorn):
            path_uvicorn="uvicorn"
    command = f"{path_uvicorn} main:app --port {port}"
    path = Path(__file__).parent
    server_process = subprocess.Popen(['cmd.exe', '/c', f'start', 'cmd', '/k', command], cwd=path)

if __name__ == "__main__":
    start_server()


def _start_server_in_same_terminal(port: int = 8000) -> None:
    """Starts the server in a separate thread."""
    def run_server():
        """Internal function to configure and run the Uvicorn server."""

        upper_folder = Path(__file__).parent
        sys.path.append(str(upper_folder))
        config = Config(app="main:app", port=port)
        server = Server(config)

        try:
            server.run()
        except Exception as e:
            print(f"Error: {e}")
    
    # Start the server in a new daemon thread
    server_thread = threading.Thread(target=run_server, daemon=True)
    server_thread.daemon = True
    server_thread.start()

    print("Server started on a separate thread.")


import json
import os
from pathlib import Path
from pprint import pprint
from typing import List

import numpy as np
from AnyQt.QtCore import QFileSystemWatcher, QObject, pyqtSignal
from AnyQt.QtGui import QImageReader
from Orange.data import ContinuousVariable, Domain, StringVariable, Table
from Orange.widgets import gui, settings, widget
from Orange.widgets.utils.signals import Output

from ..download.BaseLocalInterfaceWidget import BaseLocalInterfaceWidget


class OWLocalInterfaceImages(BaseLocalInterfaceWidget):
    name = "Local Interface - Images"
    description = ("Monitors a JSON queue file for image paths associated "
                   "with specific workflow and file IDs, and processes images.")
    icon = "icons/local_interf_img_multi_pull.svg"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/local_interf_img_multi_pull.svg"
    priority = 1215
    category = "Advanced Artificial Intelligence Tools"
    class Outputs:
        data = Output("Data", Table, default=True)

    def __init__(self):
        super().__init__()
        self.opening_method = "image_file"
        self.info_label = gui.label(self.controlArea, self, "Initial info.")       

    def process_files(self, files_to_process: list[Path]):
        images_files = [str(file_path) for file_path in files_to_process]
        table = self.create_image_table(images_files)
        self.Outputs.data.send(table)
    
    def create_image_table(self, image_files: List[str]) -> Table:
        # Define the metadata attributes for the images
        imagename_var = StringVariable.make("image name")
        imagepath_var = StringVariable.make("image path")
        imagepath_var.attributes["type"] = "image"  # Indicating type for Orange
        size_var = ContinuousVariable.make("size")
        width_var = ContinuousVariable.make("width")
        height_var = ContinuousVariable.make("height")

        domain = Domain(attributes=[], metas=[imagename_var, imagepath_var, size_var, width_var, height_var])

        meta_data = []
        
        for file_path in image_files:
            image_data = self.get_image_metadata(file_path)
            if image_data:
                meta_data.append(image_data)
        
        meta_data_array = np.array(meta_data, dtype=object)
        return Table.from_numpy(domain, X=np.empty((len(meta_data_array), 0)), metas=meta_data_array)

    def get_image_metadata(self, file_path: str):
        reader = QImageReader(file_path)
        if not reader.canRead():
            print(f"Cannot read image: {file_path}")
            return None

        img_format = bytes(reader.format()).decode("ascii")
        size = reader.size()
        width, height = (size.width(), size.height()) if size.isValid() else (float("nan"), float("nan"))
        try:
            st_size = os.stat(file_path).st_size
        except OSError:
            st_size = -1

        basename = os.path.basename(file_path)
        imgname, _ = os.path.splitext(basename)
        path_normalized = file_path.replace(os.path.sep, "/")

        return [imgname, path_normalized, st_size, width, height]

if __name__ == "__main__":
    from orangewidget.utils.widgetpreview import WidgetPreview
    WidgetPreview(OWLocalInterfaceImages).run()



from pathlib import Path
from AnyQt.QtCore import QObject, pyqtSignal, QFileSystemWatcher

class FileWatcher(QObject):
    file_modified = pyqtSignal(str)

    def __init__(self, file_path: str):
        super(FileWatcher, self).__init__()
        self.watcher = QFileSystemWatcher()
        if not Path(file_path).exists():
            Path(file_path).parent.mkdir(parents=True, exist_ok=True)
            Path(file_path).touch()
        self.watcher.addPath(file_path)
        self.watcher.fileChanged.connect(self.on_file_modified)

    def on_file_modified(self, file_path):
        self.file_modified.emit(file_path)



import os
from pathlib import Path
from pprint import pprint

from AnyQt.QtWidgets import QMessageBox
from Orange.data import Table

from ..download.BaseLocalInterfaceWidget import BaseLocalInterfaceWidget


class OWLocalInterface(BaseLocalInterfaceWidget):
    name = "Local Interface"
    description = "Get a simple data file (csv, xlsx, pkl...) from a local interface"
    icon = "icons/local_interf_pull.svg"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/local_interf_pull.svg"
    priority =1211
    category = "Advanced Artificial Intelligence Tools"
    # Specific for this widget: defines how files are opened
    opening_method = "file"

    def __init__(self):
        super().__init__()
        self.info_label.setText("Initialized Local Interface Widget.")

    def process_files(self, files_to_process: list[Path]):

        if len(files_to_process) > 1:
            print("More than one file to process. Only processing first. You may want to use MultiFile Download Widget")
            QMessageBox.critical(self, "DownloadWidget Error", "More than one file to process. Only processing first. You may want to use MultiFile Download Widget instead.")
        elif len(files_to_process) == 0:
            print("No files to process for Download Widget. Make sure you correctly set the input_id.")
            return
        file_path = files_to_process[0]

        try:
            # Load data using the defined opening method
            data_table = Table.from_file(str(file_path))
            self.info_label.setText(f"Loaded file: {str(file_path)}")
            self.Outputs.data.send(data_table)
            print(f"Data loaded successfully for workflow {self.workflow_id} and file {self.input_id}")
        except Exception as e:
            print(f"Error loading file {str(file_path)}: {e}")
            self.info_label.setText(f"Error loading file: {str(file_path)}")
        return  # Only process first matching entry



import os
from pathlib import Path
from AnyQt.QtWidgets import QLineEdit, QMessageBox, QCheckBox
from Orange.data.table import Table
from Orange.widgets import widget, gui
from Orange.widgets.widget import Input
from Orange.widgets.settings import Setting
from Orange.widgets.utils.widgetpreview import WidgetPreview
from Orange.data.io import CSVReader
import tempfile
from typing import Union
# import the orange-spectroscopy module
import json

class OWDisplayOnInterface(widget.OWWidget):
    name = "Display on local interface"
    description = "Push data to a local interface"
    icon = "icons/local_interf_push.svg"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/local_interf_push.svg"

    priority = 1210
    category = "Advanced Artificial Intelligence Tools"
    want_main_area = False
    resizing_enabled = False

    class Inputs:
        data = Input("Data", Table)

    # Persistent setting for fileId
    output_fileId: str = Setting('untitled.csv')  # type: ignore

    # Persistent setting for the Workflow id:
    workflow_id: str = Setting('test.ows') # type: ignore


    
    def __init__(self):
        super().__init__()
        self.data = None
        self.base_folder = Path(__file__).parent.parent / "webserver" / "received_files"
        self.in_process_queue = self.base_folder / "in_process_queue.json"

        if not self.in_process_queue.exists():
            self.in_process_queue.touch(exist_ok=True)
        self.setup_ui()


    def get_output_folder(self) -> Union[str, None]:
        """Returns the output folder."""
        try:
            with open(self.in_process_queue, 'r') as f:
                in_process_queue = json.load(f)

            
            for item in in_process_queue:
                if item['workflow_id'] == self.workflow_id:
                    return item['output_folder']
                
            raise ValueError("Workflow ID provided does not exist in the queue.")
        
        except json.decoder.JSONDecodeError:
            print("No items in the queue.")

        except ValueError:
            QMessageBox.critical(self, "UploadWidget Error", "Workflow ID provided does not match the workflow id of inputs.")
            


    def setup_ui(self):
        # Create a horizontal box layout for the text input
        hbox = gui.hBox(self.controlArea, "name")
        hbox2 = gui.hBox(self.controlArea, "Workflow ID")
        
        # Text input for fileId
        self.le_fileId = QLineEdit(self)
        self.le_fileId.setText(self.output_fileId)
        self.le_fileId.editingFinished.connect(self.update_fileId)
        hbox.layout().addWidget(self.le_fileId)

        # Text input for Workflow ID
        self.le_workflow_id = QLineEdit(self)
        self.le_workflow_id.setText(self.workflow_id)
        self.le_workflow_id.editingFinished.connect(self.update_workflow_id)
        hbox2.layout().addWidget(self.le_workflow_id)
        
        self.adjustSize()

    @Inputs.data
    def dataset(self, data):
        self.data = data
        if self.data is not None:
            self.save_to_file()

    def update_fileId(self):
        self.output_fileId = self.ensure_csv_extension(self.le_fileId.text())
    
    def update_workflow_id(self):   
        self.workflow_id = self.le_workflow_id.text()


    def save_to_file(self):
        if self.data is None:
            QMessageBox.warning(self, "No Data", "No data available to save.")
            return
        # Compute full output path
        output_folder = self.get_output_folder()

        if output_folder is None:
            QMessageBox.warning(self, "No file to upload.", "The queue doesn't contains any item with the provided Workflow ID.")
            return
        file_path = Path(output_folder) / self.output_fileId
        
        try:
            with tempfile.NamedTemporaryFile(delete=False) as temp:
                CSVReader.write(temp.name, self.data)
            os.replace(temp.name, file_path)
            print("Data successfully saved to: ", file_path)
        except IOError as err:
            QMessageBox.critical(self, "Error", f"Failed to save file: {err}")

    @staticmethod
    def ensure_csv_extension(fileId):
        return fileId if fileId.endswith('.csv') else fileId + '.csv'

if __name__ == "__main__":  # pragma: no cover
    WidgetPreview(OWDisplayOnInterface).run(Table("iris"))



from setuptools import setup, find_packages

# Nom du package PyPI ('pip install NAME')
NAME = "HLIT"

# Version du package PyPI
VERSION = "0.0.2"  # la version doit être supérieure à la précédente sinon la publication sera refusée

# Facultatif / Adaptable à souhait
AUTHOR = "Random but skilled people"
AUTHOR_EMAIL = ""
URL = ""
DESCRIPTION = ""
LICENSE = ""

# 'orange3 add-on' permet de rendre l'addon téléchargeable via l'interface addons d'Orange 
KEYWORDS = ["orange3 add-on",]

# Tous les packages python existants dans le projet (avec un __ini__.py)
PACKAGES = find_packages()
#PACKAGES = [pack for pack in PACKAGES if "orangecontrib" in pack]
PACKAGES = [pack for pack in PACKAGES if "orangecontrib" in pack and "OWInterface" in pack]
print(PACKAGES)


# Fichiers additionnels aux fichiers .py (comme les icons ou des .ows)
PACKAGE_DATA = {
    "orangecontrib.OWInterface.widgets": ["icons/*"],
}
# /!\ les noms de fichier 'orangecontrib.hkh_bot.widgets' doivent correspondre à l'arborescence

# Dépendances
INSTALL_REQUIRES = ["torch==2.2.2", "sentence-transformers", "gpt4all", "sacremoses", "transformers",
                    "sentencepiece", "optuna", "spacy==3.7.6", "markdown", "uvicorn", "fastapi",
                    "python-multipart"]

# Spécifie le dossier contenant les widgets et le nom de section qu'aura l'addon sur Orange
ENTRY_POINTS = {
    "orange.widgets": (
        "OWInterface = orangecontrib.OWInterface.widgets",
    )
}

NAMESPACE_PACKAGES = ["orangecontrib"]

setup(name=NAME,
      version=VERSION,
      author=AUTHOR,
      author_email=AUTHOR_EMAIL,
      url=URL,
      description=DESCRIPTION,
      license=LICENSE,
      keywords=KEYWORDS,
      packages=PACKAGES,
      package_data=PACKAGE_DATA,
      install_requires=INSTALL_REQUIRES,
      entry_points=ENTRY_POINTS,
      namespace_packages=NAMESPACE_PACKAGES,
      )



import json
import os
from itertools import chain, repeat
from pathlib import Path
from pprint import pprint
from typing import Union

import numpy as np
from AnyQt.QtCore import QFileSystemWatcher, QObject, pyqtSignal
from Orange.data import Domain, StringVariable, Table
from Orange.widgets import gui, settings, widget

from ..download.BaseLocalInterfaceWidget import BaseLocalInterfaceWidget


class OWLocalInterfaceMultifile(BaseLocalInterfaceWidget):
    name = "Local Interface - Multifile"
    description = "Get multiple data file (csv, xlsx, pkl...) from a local interface"
    icon = "icons/local_interf_multi_pull.svg"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/local_interf_multi_pull.svg"
    priority = 1213
    category = "Advanced Artificial Intelligence Tools"
    class Outputs:
        data = widget.Output("Data", Table)

    def __init__(self):
        super().__init__()
        self.opening_method = "multiple_file"

    def process_files(self, files_to_process: list[Path]):
        data = self.concatenate_data_with_origin(files_to_process)
        self.Outputs.data.send(data)

    def concatenate_data_with_origin(self, filepaths: list[Path])->Union[Table,None]:
        if not filepaths:
            return None
        
        filenames = [filepath.name for filepath in filepaths]
        tables = [Table.from_file(str(x)) for x in filepaths]

        # Determine the merged domain, accounting for all attributes and metas in all tables
        domain = self._merge_domains([table.domain for table in tables])
        source_var = StringVariable("File Source")
        domain = Domain(domain.attributes, domain.class_vars, domain.metas + (source_var,))

        # Transform tables to the new domain so they can be concatenated
        tables = [table.transform(domain) for table in tables]
        data = Table.concatenate(tables)

        with data.unlocked():
            # Flatten the array to the correct shape
            data.metas[:, -1] = np.array(list(
                chain(*(repeat(fn, len(table))
                        for fn, table in zip(filenames, tables)))
            ))

        return data

    def _merge_domains(self, domains):
        # This assumes all tables have compatible domains; more complex logic can be implemented if needed
        return domains[0] if domains else Domain([])

if __name__ == "__main__":
    # This would normally start the GUI application
    pass



import asyncio
import os
import time
from contextlib import asynccontextmanager
from pathlib import Path
from typing import Any, Dict, List

from fastapi import FastAPI, File, HTTPException, UploadFile
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import FileResponse, JSONResponse

from fileQueue import FileQueue
from utils import (generate_unique_output_folder, hash_file_content, make_list,
                   zip_folder)


@asynccontextmanager
async def lifespan(app: FastAPI):
    os.makedirs(save_directory, exist_ok=True)
    os.makedirs(output_directory, exist_ok=True)
    asyncio.create_task(process_files_queue())
    yield

app = FastAPI(lifespan=lifespan)
# CORS Settings
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],  # List of origins that are allowed, '*' means allow all origins
    allow_credentials=True,
    allow_methods=["*"],  # List of HTTP methods that are allowed, '*' means allow all methods
    allow_headers=["*"],  # List of headers that are allowed, '*' means allow all headers
)

# Constants
MAX_PROCESSING_QUEUE_LENGTH: int = 1
TIMEOUT_LIMIT: int = 120  # seconds

# Path settingsmatch_number
save_directory: Path = Path(__file__).parent / "received_files"
output_directory: Path = Path(__file__).parent / "sent_files"

print("save_directory: ", save_directory)
print("output_directory: ", output_directory)

waiting_queue: FileQueue = FileQueue(save_directory / "waiting_queue.json")
in_process_queue: FileQueue = FileQueue(save_directory / "in_process_queue.json")

def make_tuple_list(text_inputs: List[str]) -> List[tuple[str, str]]:
    assert len(text_inputs) % 2 == 0, "The number of text inputs must be even."
    print("text_inputs: ", text_inputs)
    print("return: ", [(text_inputs[i], text_inputs[i+1]) for i in range(0, len(text_inputs), 2)])
    return [(text_inputs[i], text_inputs[i+1]) for i in range(0, len(text_inputs), 2)]

# Main API Endpoints
@app.post("/upload-files/")
async def upload_files(
    file_ids: List[str], 
    workflow_id: str,
    opening_methods: List[str], 
    expected_nb_outputs: int, 
    files: List[UploadFile] = File(...),
    text_inputs: List[str] = []
) -> JSONResponse:
    """
    Upload multiple files to the server for processing through a Orange workflow. 
      
    Args:   
        - file_ids (List[str]): List of unique file IDs corresponding to the files.
          The IDs should match the ones configured in the Orange workflow.   
        - workflow_id (str): Unique ID of the workflow.
          The ID should match the one configured in the Orange workflow.   
        - opening_methods (List[str]): List of opening methods for the files.
          Choose from 'image_file', 'multiple_file', or 'file'.    
        - expected_nb_outputs (int): Expected number of output files (depending on your workflow).
        - files (List[UploadFile]): List of files to be uploaded.  
        - text_inputs (List[str]): List of text inputs for the workflow. Must be in the format [input_id1, value1, input_id2, value2].

    Returns:   
        JSONResponse: Response message with the unique ID of the processing task."""
    text_inputs = make_list(text_inputs)
    opening_methods = make_list(opening_methods)
    print(file_ids, opening_methods)
    if len(files) != len(file_ids) or len(files) != len(opening_methods):
        raise HTTPException(status_code=400, detail="Number of files, file IDs, and opening methods must match.")

    file_entries: List[Dict[str, Any]] = []
    for file, file_id, opening_method in zip(files, file_ids, opening_methods):
        if opening_method not in ["image_file", "multiple_file", "file"]:
            raise HTTPException(status_code=400, detail="Invalid opening method provided. Choose from 'image_file', 'multiple_file', or 'file'.")
        
        content: bytes = await file.read()
        file_hash: str = hash_file_content(content)
        input_filename: str = str(save_directory / str(file.filename))
        
        # Save the file locally
        with open(input_filename, 'wb') as f:
            f.write(content)

        file_entries.append({
            'input_filename': input_filename,
            'file_id': file_id,
            'file_hash': file_hash,
            'opening_method': opening_method,
        })

    unique_id, output_folder = generate_unique_output_folder(output_directory)

    waiting_queue.add({
        'unique_id': unique_id,
        'workflow_id': workflow_id,
        'file_entries': file_entries,
        'output_folder': str(output_folder),
        'expected_nb_outputs': expected_nb_outputs,
        'text_inputs': text_inputs
    })

    return JSONResponse(content={"message": "Files queued for processing", "unique_id": unique_id})

@app.post("/upload-files-via-path/")
async def upload_files_via_path(
    filepaths: List[str], 
    file_ids: List[str], 
    opening_methods: List[str],
    workflow_id: str, 
    expected_nb_outputs: int,
    text_inputs: List[tuple[str, str]] = []
) -> JSONResponse:
    """
    Upload multiple files to the server for processing through a Orange workflow.

    Args:   
        - filepaths (List[str]): List of file paths to be uploaded.   
        - file_ids (List[str]): List of unique file IDs corresponding to the files.
          The IDs should match the ones configured in the Orange workflow.  
        - opening_methods (List[str]): List of opening methods for the files.
          Choose from 'image_file', 'multiple_file', or 'file'.   
        - workflow_id (str): Unique ID of the workflow.
          The ID should match the one configured in the Orange workflow.   
        - expected_nb_outputs (int): Expected number of output files (depending on your workflow).

    Returns:   
        JSONResponse: Response message with the unique ID of the processing task."""
    file_ids = make_list(file_ids)
    opening_methods = make_list(opening_methods)
    filepaths = make_list(filepaths)
    if len(filepaths) != len(file_ids) or len(file_ids) != len(opening_methods):
        raise HTTPException(status_code=400, detail="Number of file paths, file IDs, and opening methods must match.")

    file_entries: List[Dict[str, Any]] = []
    for filepath, file_id, opening_method in zip(filepaths, file_ids, opening_methods):
        if opening_method not in ["image_file", "multiple_file", "file"]:
            raise HTTPException(status_code=400, detail="Invalid opening method provided.")
        
        file_path: Path = Path(filepath)
        if not file_path.exists():
            raise HTTPException(status_code=404, detail=f"File not found: {filepath}")

        with file_path.open('rb') as f:
            content: bytes = f.read()

        file_hash: str = hash_file_content(content)
        file_entries.append({
            'input_filename': str(file_path),
            'file_id': file_id,
            'file_hash': file_hash,
            'opening_method': opening_method
        })

    unique_id, output_folder = generate_unique_output_folder(output_directory)

    waiting_queue.add({
        'unique_id': unique_id,
        'workflow_id': workflow_id,
        'file_entries': file_entries,
        'output_folder': str(output_folder),
        'expected_nb_outputs': expected_nb_outputs,
        'text_inputs': text_inputs
    })

    return JSONResponse(content={"message": "Files queued for processing", "unique_id": unique_id})

@app.get("/retrieve-outputs/{unique_id}", response_model=None)
async def retrieve_outputs(unique_id: str, file_id: str = "") -> FileResponse | JSONResponse:
    """
    Retrieve the output files for a given unique ID. By default, a zip of all files will be returned.
    If a specific output filename is provided, only that file will be returned.   
    **IMPORTANT** : If the processing is not complete, the method will return a 404 error. You have to keep calling this method until the processing is complete.

    Args:     
        - unique_id (str): Unique ID of the processing task.  
        - file_id (str): Optional file ID to retrieve a specific file. If not provided, a zip of all files will be returned. 

    Returns:  
        FileResponse | JSONResponse: Response with the output file(s) for download.
    """
    output_folder = output_directory / unique_id
    
    # Check if the processing is complete by checking the presence of the "done" file
    if (output_folder / "done").exists():
        all_files = list(output_folder.glob('*'))
        
        # Remove 'done' and 'output.zip' files from the list
        for file in [output_folder / "done", output_folder / "output.zip"]:
            if file.exists():
                all_files.remove(file)

        # If a specific file_id is provided, search for it in the output folder
        if file_id:
            for file in all_files:
                if file_id in file.name:
                    return FileResponse(str(file), media_type='application/octet-stream', filename=file.name)
            return JSONResponse(content={"message": f"File with ID {file_id} not found."}, status_code=404)

        # If no specific file_id is specified, handle the retrieval of multiple or single files
        if len(all_files) == 1:
            # Return the single file if only one file is present
            output_file = all_files[0]
            return FileResponse(str(output_file), media_type='application/octet-stream', filename=output_file.name)
        elif len(all_files) > 1:
            # Create a zip of all the output files for download
            output_zip: Path = output_folder / "output.zip"
            zip_folder(output_folder, output_zip)
            return FileResponse(str(output_zip), media_type='application/zip', filename=f"output.zip")
        else:
            return JSONResponse(content={"message": "No files found in the output folder."}, status_code=404)

    elif output_folder.exists():
        return JSONResponse(content={"message": "Processing is not complete."}, status_code=404)
    else:
        return JSONResponse(content={"message": "Unique ID not found."}, status_code=404)

@app.get("/ping")
async def ping() -> JSONResponse:
    return JSONResponse(content={"message": "pong"})



# Background Processing Task
async def process_files_queue() -> None:
    """
    Background task to process files from the waiting queue.
    This task will check the waiting queue for new files and move them to the in-process queue.
    It will then wait for the expected number of output files to be generated by the workflow.
    """

    while True:
        if not waiting_queue.queue or len(in_process_queue.queue) >= MAX_PROCESSING_QUEUE_LENGTH:
            await asyncio.sleep(1)
            continue

        workflow_entry= waiting_queue.queue[0] # element en attente
        waiting_queue.remove(workflow_entry)
        in_process_queue.add(workflow_entry) # element en process
        expected_nb_outputs: int = workflow_entry['expected_nb_outputs']
        output_folder: Path = Path(workflow_entry['output_folder'])
        
        # Wait for the expected number of output files to be generated by the workflow
        start_time: float = time.time()
        output_folder: Path = Path(workflow_entry['output_folder'])
        while time.time() - start_time < TIMEOUT_LIMIT:
            if len(list(output_folder.glob('*'))) == expected_nb_outputs:
                break
            elif len(list(output_folder.glob('*'))) > expected_nb_outputs:
                raise Exception("Too many files found in the output folder. Error.")
            await asyncio.sleep(1)
        
        in_process_queue.remove(workflow_entry)
        (output_folder / "done").touch()


